# Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import colorsys
from functools import partial
from PIL import Image,ImageFilter
import io
import multiprocessing
import subprocess
import sys

VERSION = '0.1'

class EncodeRle:
    def __init__(self,rlmin=2,rlmax=255):
        self.rlmin = rlmin
        self.rlmax = rlmax
        self.dst = bytearray()
        self.count = 0
        self.prev = None
        self.buf = bytearray()

    def flush(self):
        if self.count >= self.rlmin:
            self.add_runlength(self.prev,self.count)
        else:
            for i in range(self.count):
                self.add_single(self.prev)

    def encode(self,src):
        for b in src:
            if b == self.prev:
                self.count+= 1
                if self.count >= self.rlmax:
                    self.flush()
                    self.count = 0
            else:
                self.flush()
                self.count = 1
                self.prev = b
        self.flush()
        return self.dst

class EncodeKoala(EncodeRle):
    def __init__(self):
        super(EncodeKoala,self).__init__(4)

    def add_single(self,b):
        if 0xfe == b:
            self.dst.append(0xfe)
            self.dst.append(b)
            self.dst.append(1)
        else:
            self.dst.append(b)

    def add_runlength(self,b,count):
        self.dst.append(0xfe)
        self.dst.append(b)
        self.dst.append(count)

def quant(img,colors):
    p = subprocess.Popen(['pyriel-quant','--colors',str(colors)],stdin=subprocess.PIPE,stdout=subprocess.PIPE)
    img.save(p.stdin,format='PNG')
    data, err = p.communicate()
    if not err:
        img = Image.open(io.BytesIO(data)).convert('RGB')
    else:
        print('error:',err)
        exit(1)
    palimg = img.convert('P',palette=Image.ADAPTIVE,colors=colors)
    palimg.load()
    return palimg.palette.tobytes()[:colors*3]

def remap(img,palette):
    colors = int(len(palette) / 3)
    palimg = Image.new('P',(1,1))
    palimg.putpalette(palette * int(256/colors))
    return img.convert('RGB').quantize(colors=colors,palette=palimg)

class Vram:
    def from_pixels_mono(self,pixels):
        stride = self.width >> 3
        for y in range(self.height):
            for x in range(stride):
                b = 0
                for i in range(8):
                    b<<= 1
                    if pixels[x*8+i,y] > 0:
                        b|= 1
                self.mem[x + y * stride] = b
        return self

    def from_image_mono(self,img):
        return self.from_pixels_mono(img.resize((self.width,self.height)).convert('1').load())

    def from_image(self,img):
        return self.from_pixels(img.load())

    def from_png(self,data):
        return self.from_image(Image.open(io.BytesIO(data)))

class AmstradCpc(Vram):
    def __init__(self,mode):
        self.mode = mode
        if 0 == mode:
            self.width = 160
            self.height = 200
        elif 1 == mode:
            self.width = 320
            self.height = 200
        elif 2 == mode:
            self.width = 640
            self.height = 200

    def from_image(self,img):
        if 0 == self.mode:
            return
        if 1 == self.mode:
            return
        if 2 == self.mode:
            return from_image_mono(img)

class VramApple2(Vram):
    def __init__(self):
        self.width = 280
        self.height = 192
        self.mem = bytearray(6720)

    def from_image(self,img):
        return self.from_image_mono(img)

class VramAtari8bit(Vram):
    def __init__(self,mode):
        self.mode = mode
        if 3 == mode:
            self.width = 40
            self.height = 24
            self.colors = 4
        elif 4 == mode:
            self.width = 80
            self.height = 48
            self.colors = 2
        elif 5 == mode:
            self.width = 80
            self.height = 48
            self.colors = 4
        elif 6 == mode:
            self.width = 160
            self.height = 96
            self.colors = 2
        elif 7 == mode:
            self.width = 160
            self.height = 96
            self.colors = 4
        elif 8 == mode:
            self.width = 320
            self.height = 192
            self.colors = 2
        elif 9 == mode:
            self.width = 80
            self.height = 192
            self.colors = 16
        elif 10:
            self.width = 80
            self.height = 192
            self.colors = 9
        elif 11:
            self.width = 80
            self.height = 192
            self.colors = 16
        elif 14:
            self.width = 160
            self.height = 192
            self.colors = 2
        elif 15:
            self.width = 160
            self.height = 192
            self.colors = 4


        self.mem = bytearray(7680)
        self.palette = bytearray(16)
        self.default_palette = bytes([0,    # Black
                                      16,   # Rust
                                      32,   # Red-orange
                                      48,   # Dark orange
                                      64,   # Red
                                      80,   # Dark lavender
                                      96,   # Cobalt blue
                                      112,  # Ultramarine
                                      128,  # Medium blue
                                      144,  # Dark blue
                                      160,  # Blue-grey
                                      176,  # Olive green
                                      192,  # Medium green
                                      208,  # Dark green
                                      224,  # Orange-green
                                      240]) # Orange
        self.palette[:] = self.default_palette

    def from_pixels(self,pixels):
        if 2 == self.colors:
            return
        elif 4 == self.colors:
            stride = self.width >> 2
            for y in range(self.height):
                for x in range(stride):
                    b = 0
                    for i in range(4):
                        b<<= 2
                        b|= pixels[x*4+i,y] & 3
                    self.mem[x + y * stride] = b

    def from_image(self,img):
        if 2 == self.colors:
            return from_image_mono(img)
        else:
            #pal = ImagePalette.ImagePalette(size=self.colors)
            img = img.resize((self.width,self.height))
            img = img.quantize(self.colors)
            pal = img.getpalette()
            for i in range(self.colors):
                #r,g,b = pal[i]
                r = pal[i*3+0]
                g = pal[i*3+1]
                b = pal[i*3+2]
                h,s,v = colorsys.rgb_to_hsv(r/255.,g/255.,b/255.)
                self.palette[i] = int(h*16)<<4 | int(v*8)<<1
            return super(VramAtari8bit,self).from_image(img)

class VramAtarist(Vram):
    def __init__(self,res,width,height):
        self.width = width
        self.height = height
        self.resolution = bytes([0,res])
        self.screen = bytearray(32000)
        self.palette = bytearray(32)

class VramAtaristLow(VramAtarist):
    def __init__(self,width=320,height=200):
        super(VramAtaristLow,self).__init__(0,width,height)

    def from_image(self,img,width=320,height=200):
        if width > 0 and height > 0:
            img = img.resize((width,height))
        pal = quant(img,16)
        img = remap(img,pal)
        pixels = img.load()
        for y in range(height):
            for x in range(int(width/16)):
                for plane in range(4):
                    b0 = 0
                    b1 = 0
                    for i in range(8):
                        if pixels[x*16+0+i,y] & (1<<plane):
                            b0 |= 128>>i
                        if pixels[x*16+8+i,y] & (1<<plane):
                            b1 |= 128>>i
                    self.screen[y*160 + x*8+plane*2+0] = b0
                    self.screen[y*160 + x*8+plane*2+1] = b1

        for i in range(16):
            r = pal[i*3+0] >> 5
            g = pal[i*3+1] >> 5
            b = pal[i*3+2] >> 5
            rgb = r<<8|g<<4|b
            self.palette[i*2+0] = rgb >> 8
            self.palette[i*2+1] = rgb & 255
        return self

    def from_pi1(self,data):
        self.palette[:] = data[2:][:32]
        self.screen[:] = data[2+32:][:32000]
        return self

class VramAtaristMed(VramAtarist):
    def __init__(self,width=640,height=200):
        super(VramAtaristMed,self).__init__(1,width,height)


    def from_pi2(self,data):
        self.palette[:] = data[2:][:32]
        self.screen[:] = data[2+32:][:32000]
        return self

class VramAtaristHigh(VramAtarist):
    def __init__(self,width=640,height=400):
        super(VramAtaristHigh,self).__init__(2,width,height)

    def from_pixels_mono(self,pixels):
        stride = self.width >> 3
        for y in range(self.height):
            for x in range(stride):
                b = 0
                for i in range(8):
                    b<<= 1
                    if pixels[x*8+i,y] <= 0:
                        b|= 1
                self.screen[x + y * stride] = b
        return self

    def from_image(self,img):
        return self.from_image_mono(img)

    def from_pi3(self,data):
        self.screen[:] = data[2+32:][:32000]
        return self

class VramC64(Vram):
    def __init__(self):
        self.bitmap = bytearray(8000)
        self.screen = bytearray(1000)
        self.color = bytearray(1000)
        self.background = 0

class VramC64Highres(VramC64):
    def from_pixels(self,pixels):
        pal = [0]*2
        for ty in range(25):
            for tx in range(40):
                col = [self.background]
                for y in range(8):
                    row = 0
                    for x in range(8):
                        pix = pixels[tx*8+x,ty*8+y]
                        if not pix in col:
                            if len(col) >= 2:
                                print('Too many colors in cell {},{}'.format(tx,ty))
                                return False
                            col.append(pix)
                        v = col.index(pix)
                        pal[v] = pix
                        row <<= 1
                        row |= v
                    self.bitmap[y+(tx+ty*40)*8] = row
                self.screen[tx+ty*40] = pal[1]<<4 | pal[0]
        return True

class VramC64Multicolor(VramC64):
    def from_pixels(self,pixels):
        pal = [0]*4
        for ty in range(25):
            for tx in range(40):
                col = [self.background]
                for y in range(8):
                    row = 0
                    for x in range(4):
                        pix = pixels[tx*4+x,ty*8+y]
                        if not pix in col:
                            if len(col) >= 4:
                                print('Too many colors in cell {},{}'.format(tx,ty))
                                return False
                            col.append(pix)
                        v = col.index(pix)
                        pal[v] = pix
                        row <<= 2
                        row |= v
                    self.bitmap[y+(tx+ty*40)*8] = row
                self.screen[tx+ty*40] = pal[1]<<4 | pal[2]
                self.color[tx+ty*40] = pal[3]
        return True

    def from_koala(self,data):
        self.bitmap = data[2:8002]
        self.screen = data[8002:9002]
        self.color = data[9002:10002]
        self.background = data[10002]
        return self

class VramPc88(Vram):
    def __init__(self):
        self.mem = bytearray(48000)
        self.palette = bytes(   0,   0,   0,
                             0xff,   0,   0,
                                0,0xff,   0,
                             0xff,0xff,   0,
                                0,   0,0xff,
                             0xff,   0,0xff,
                                0,0xff,0xff,
                             0xff,0xff,0xff)

    def from_image(self,img):
        img = img.resize((self.width,self.height))
        #img = img.convert(
        img = img.quantize(colors=8,palette=self.palette)
        return self.from_pixels(img.load())

class VramPc88va(Vram):
    def __init__(self):
        self.mem = bytearray(256000)

class VramPc98(Vram):
    def __init__(self):
        self.mem = bytearray(128000)

class VramSpectrum(Vram):
    palette = [   0,   0,   0,
                  0,   0,0xd7,
               0xd7,   0,   0,
               0xd7,   0,0xd7,
                  0,0xd7,   0,
                  0,0xd7,0xd7,
               0xd7,0xd7,   0,
               0xd7,0xd7,0xd7,
                  0,   0,   0,
                  0,   0,0xff,
               0xff,   0,   0,
               0xff,   0,0xff,
                  0,0xff,   0,
                  0,0xff,0xff,
               0xff,0xff,   0,
               0xff,0xff,0xff]
    def __init__(self,tileheight=1):
        self.tileheight = tileheight
        self.mem = bytearray(6144)
        self.attributes = bytearray(6144)

    def calc_y(self,y):
        return (y & 192) | (y & 7)<<3 | (y & 56)>>3

    def from_pixels(self,pixels):
        offs = 0
        for ty in range(192//self.tileheight):
            for tx in range(32):
                colors = []
                for y1 in range(self.tileheight):
                    b = 0
                    for x in range(8):
                        #p = pixels[tx*8+x,self.calc_y(ty*self.tileheight + y1)]
                        p = pixels[tx*8+x,ty*self.tileheight + y1]
                        #print('pixel',p)
                        if p in colors:
                            p = colors.index(p)
                        elif len(colors) < 2:
                            colors.append(p)
                            p = len(colors)-1
                        else:
                            # Too many colors
                            print('too many colors in tile {}x{}'.format(tx,ty))
                            p = 0
                        b<<= 1
                        if p <= 0:
                            b|= 1
                    self.mem[tx + self.calc_y(ty*self.tileheight + y1)*32] = b
                    offs+= 1
                while len(colors) < 2:
                    colors.append(0)
                bright = 1 if colors[0] >= 8 or colors[1] >= 8 else 0
                attr = bright<<6 | (colors[0]&7) | (colors[1]&7)<<3
                self.attributes[ty*32+tx] = attr
        return self

    def from_pixels_mask(self,pixels,mask):
        offs = 0
        for y1 in range(192):
            for tx in range(32):
                b = 0
                for x in range(8):
                    b<<= 1
                    if pixels[tx*8+x,self.calc_y(y1)] & mask > 0:
                        b|= 1
                self.mem[offs] = b
                offs+= 1
        return self

    def from_image(self,img):
        p = subprocess.Popen(['pyriel-spectrum','--tileheight',str(self.tileheight)],stdin=subprocess.PIPE,stdout=subprocess.PIPE)
        img.save(p.stdin,format='PNG')
        data, err = p.communicate()
        if err:
            print('error:',err)
            exit(1)
        img = Image.open(io.BytesIO(data)).convert('RGB')
        return self.from_pixels(remap(img,self.palette).load())

class Conv:
    def __init__(self,inext,outext,inname=None,outname=None):
        self.out = None
        if not inname:
            inname = inext
        if not outname:
            outname = outext
        parser = argparse.ArgumentParser(description='Convert a {} image into a {} image.'.format(inname,outname))
        parser.add_argument('infile', metavar='<input.{}>'.format(inext), type=str, nargs='?', help='input file (use - or leave empty for stdin)', default='-')
        parser.add_argument('outfile', metavar='<output.{}>'.format(outext), type=str, nargs='?', help='output file (use - or leave empty for stdout)', default='-')
        parser.add_argument('-V', '--version', action='version', version=VERSION, help='show program\'s version number ({}) and exit'.format(VERSION))
        args = parser.parse_args()

        if '-'==args.infile:
            self.data = sys.stdin.buffer.read()
        else:
            with open(args.infile,'rb') as f:
                self.data = f.read()
        if '-'==args.outfile:
            self.out = sys.stdout.buffer
        else:
            self.out = open(args.outfile,"wb")

    def __del__(self):
        if self.out:
            self.out.close()

    def write(self,data):
        self.out.write(data)
