/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <limits.h>                     /* INT_MAX */
#include <string.h>                     /* memcpy() */

#include "common.h"
#include "convert.h"
#include "debug.h"
#include "nile.h"
#include "nil-blit.h"
#include "nil-blur.h"
#include "nil-c64.h"
#include "nil-png-r.h"
#include "nil-png-w.h"
#include "nil-quant.h"
#include "nil-remap.h"
#include "nil-resize.h"
#include "ni-tiled.h"

#define C64_WIDTH 160
#define C64_HEIGHT 200


static nil_dither_type_t opt_dither = NIL_DITHER_ORD8;
static nil_c64_palette_t opt_palette = NIL_C64_COLODORE;


static void quant_image(nile_t *ni, FILE *f){
    nile_t screen[16];
    double matrix[5*5];
    int best_diff = INT_MAX;
    int best_bg = 0;
    int bg;

    nil_blur_gaussian_matrix(matrix, 5, 1);

    nile_t org;
    nile_t tile;
    nile_init(&org);
    org.palette = ni->palette;
    nile_init(&tile);
    for(bg=0;bg<16;++bg){
        nile_t *dst = nile_init(&screen[bg]);
        int diff;
        int ty;
        dst->colors = ni->colors;
        //nile_allocate_palette(dst, 16, ni->bpc);
        dst->format = ni->format;
        //nile_copy_colors_rgb(dst->palette, dst->bpc, dst->format, ni->palette, ni
        dst->palette = ni->palette;
        nile_ensure_indexed(dst, C64_WIDTH, C64_HEIGHT, ni->bpc);
        nile_ensure_planar(dst, C64_WIDTH, C64_HEIGHT, ni->bpc);
        nil_paste_planar(dst, ni, 0, 0);
        for(ty=0;ty<C64_HEIGHT/8;++ty){
            int tx;
            for(tx=0;tx<C64_WIDTH/4;++tx){
                uint8_t remap_table[256];
                nil_quant_set_t set;
                int i;
                ni_excerpt_indexed(&org, ni, tx*4, ty*8, 4, 8);
                ni_excerpt_planar(&tile, dst, tx*4, ty*8, 4, 8);
                nil_quant_set_init(&set, 8);
                for(i=0;i<ni->colors;++i)
                    nil_quant_set_add(&set, ni->palette + i*ni->bpc, i==bg?-1:0, ni);
                nil_quant_set_add_nearest_indexed_pixels(&set, &org);
                nil_quant_set_reduce_simple(&set, 4, ni);
                nil_quant_set_get_palette(&set, &tile);
                dither(&tile, opt_dither, tx*8, 0, 4);
                nil_remap_create_table(remap_table, dst, &tile);
                nil_remap_apply_table(remap_table, &tile);
                nil_paste_indexed(dst, &tile, tx*4, ty*8);
                nil_quant_set_done(&set);
                org.indexed = NULL;
                tile.planar = NULL;
            }
        }
        nil_blur_apply_matrix_indexed(dst, matrix, 5);
        diff = compare_planar(ni, dst);
        if(diff < best_diff){
            best_diff = diff;
            best_bg = bg;
        }
        //nile_done(dst);
    }
    DEBUG_p("the best background color is %d",best_bg);

    nile_drop_planar(&screen[best_bg]);
    nil_png_write(&screen[best_bg],f);

    org.palette = NULL;
    nile_done(&org);
    nile_done(&tile);
}


static int on_option(int opt, const char *arg){
    switch(opt){
    DITHER_PARSE
    C64_PALETTE_PARSE
    default:
        return -1;
    }
}


int main(int argc, char **argv){
    static const convert_option_t options[] = {
        DITHER_OPTIONS
        C64_PALETTE_OPTIONS
        {NULL}
    };
    convert_t conv;
    nile_t src;
    nile_t dst;

    convert_init(&conv, argc, argv, "png", "png", NULL, options, on_option);
    nil_png_read(&src, conv.in);

    nil_resize(&dst, &src, C64_WIDTH, C64_HEIGHT);
    //nil_png_save(&dst, "tmp-in.png");
    nil_c64_set_palette(&dst, opt_palette);
    dither(&dst, opt_dither, 0, 0, 4);
    //nil_png_save(&dst, "tmp-ideal.png");
    quant_image(&dst, conv.out);

    convert_done(&conv);
    return 0;
}
