/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <float.h>
#include <limits.h>                     /* INT_MAX */
#include <stdio.h>
#include <stdlib.h>                     /* atoi() */
#include <string.h>                     /* memcpy() */

#include "common.h"
#include "convert.h"
#include "debug.h"
#include "nil-blit.h"
#include "nil-blur.h"
#include "nil-dither.h"
#include "nil-png-r.h"
#include "nil-png-w.h"
#include "nil-resize.h"
#include "nil-spectrum.h"
#include "ni-tiled.h"

#define SPECTRUM_WIDTH 256
#define SPECTRUM_HEIGHT 192

#define COLOR_COMBINATIONS 56

#ifndef ABS
# define ABS(a) ((a)<0?-(a):(a))
#endif


static const uint8_t spectrum_palette[] = {
       0,   0,   0,
       0,   0,0xd7,
    0xd7,   0,   0,
    0xd7,   0,0xd7,
       0,0xd7,   0,
       0,0xd7,0xd7,
    0xd7,0xd7,   0,
    0xd7,0xd7,0xd7,
       0,   0,   0,
       0,   0,0xff,
    0xff,   0,   0,
    0xff,   0,0xff,
       0,0xff,   0,
       0,0xff,0xff,
    0xff,0xff,   0,
    0xff,0xff,0xff
};

static const int color_combination[COLOR_COMBINATIONS * 2] = {
     0,  1,  0,  2,  0,  3,  0,  4,  0,  5,  0,  6,  0,  7,
             1,  2,  1,  3,  1,  4,  1,  5,  1,  6,  1,  7,
                     2,  3,  2,  4,  2,  5,  2,  6,  2,  7,
                             3,  4,  3,  5,  3,  6,  3,  7,
                                     4,  5,  4,  6,  4,  7,
                                             5,  6,  5,  7,
                                                     6,  7,
             8, 10,  8, 11,  8, 12,  8, 13,  8, 14,  8, 15,
             9, 10,  9, 11,  9, 12,  9, 13,  9, 14,  9, 15,
                    10, 11, 10, 12, 10, 13, 10, 14, 10, 15,
                            11, 12, 11, 13, 11, 14, 11, 15,
                                    12, 13, 12, 14, 12, 15,
                                            13, 14, 13, 15,
                                                    14, 15
};


static nil_dither_type_t opt_dither = NIL_DITHER_JJN;
static int opt_tileheight = 8;


static void setup_tiles(nile_t *ni, int height, int bpc, nile_format_t format){
    int i;
    for(i=0;i<COLOR_COMBINATIONS;++i){
        nile_init(&ni[i]);
        nile_ensure_palette(&ni[i],2,bpc);
        nile_ensure_indexed(&ni[i],8,height,2);
        nile_ensure_planar(&ni[i],8,height,bpc);
        ni[i].format = format;
        nile_copy_colors_rgb(ni[i].palette + bpc*0,bpc,format,spectrum_palette + color_combination[i*2+0]*3, 3, NILE_FORMAT_r8g8b8, 1);
        nile_copy_colors_rgb(ni[i].palette + bpc*1,bpc,format,spectrum_palette + color_combination[i*2+1]*3, 3, NILE_FORMAT_r8g8b8, 1);
    }
}


#if 0
static int compare_planar_to_indexed(nile_t *ni1, nile_t *ni2, int segwidth, int segheight){
    int diff = 0;
    const uint8_t *p1 = ni1->planar;
    int stride1 = nile_get_planar_stride(ni1);
    int skip1 = stride1 - ni1->width * ni1->bytes_per_color;
    const uint8_t *p2 = ni2->indexed;
    int stride2 = nile_get_indexed_stride(ni2);
    int skip2 = stride2 - ni2->width;
    int sy;
    for(sy=0;sy<ni1->height;sy+=segheight){
        int sx;
        for(sx=0;sx<ni1->width;sx+=segwidth){
            int subdiff[4];
            int y;
            int i;
            for(i=0;i<ni1->bytes_per_color;++i)
                subdiff[i] = 0;
            for(y=0;y<segheight;++y){
                int x;
                for(x=0;x<segwidth;++x){
                    const uint8_t *p1 = ni1->planar + (sx+x)*ni1->bytes_per_color + (sy+y)*stride1;
                    const uint8_t *p2 = ni2->indexed + (sx+x) + (sy+y)*stride2;
                    for(i=0;i<ni1->bytes_per_color;++i)
                        subdiff[i] += p1[i] - ni2->palette[*p2 * ni2->bytes_per_color + i];
                        //diff += ABS(p1[i] - ni2->palette[*p2 * ni2->bytes_per_color + i]);
                }
            }
            for(i=0;i<ni1->bytes_per_color;++i)
                diff += ABS(subdiff[i]);
            //p1 += ni1->bytes_per_color;
            //++p2;
        }
        //p1 += skip1;
        //p2 += skip2;
    }
    return diff;
}
#endif


static void copy_to_bitplane(uint8_t *dst, const uint8_t *src, int mask, int count){
    int i;
    for(i=0;i<count;++i){
        uint8_t b = 0;
        if((i&7) == 0){
            b = 0;
        }else{
            b <<= 1;
        }
        if(*src++ & mask)
            b |= 1;
        if((i&7) == 7)
            *dst++ = b;
    }
}


static void paste_tile(nil_spectrum_vram_t *vram, nile_t *tile, int bg, int fg, int px, int py){
    int stride = nile_get_indexed_stride(tile);
    int y;
    for(y=0;y<tile->height;++y)
        copy_to_bitplane(vram->bitmap + px/8 + nil_spectrum_translate_y(py+y)*32, tile->indexed + y*stride, 1, tile->width);
    vram->attributes[px/tile->width + (py/tile->height)*32] = (fg>=8||bg>=8?1<<6:0)|(fg&7)<<3|(bg&7);
}


static void quant_image(nile_t *ni,nil_spectrum_vram_t *vram,int tileheight){
    nile_t tile[COLOR_COMBINATIONS];
    double matrix[5*5];
    int ty;

    setup_tiles(tile, tileheight, ni->bpc, ni->format);
    nil_blur_gaussian_matrix(matrix,5,0.9);

    for(ty=0;ty<SPECTRUM_HEIGHT/tileheight;++ty){
        int tx;
        DEBUG_p("%d",ty);
        for(tx=0;tx<SPECTRUM_WIDTH/8;++tx){
            nile_t org;
            double best_value = DBL_MAX;
            int best_index = 0;
            nile_t *best_tile = NULL;
            int i;
            nile_init(&org);
            ni_excerpt_planar(&org, ni, tx*8, ty*tileheight, 8, tileheight);
            for(i=0;i<COLOR_COMBINATIONS;++i){
                int value;
                nil_paste_planar(&tile[i],&org,0,0);
                dither(&tile[i], opt_dither, tx*8, ty*tileheight, 2);
                nil_blur_apply_matrix_indexed(&tile[i],matrix,5);
#if 1
                value = compare_planar(&org,&tile[i]);
#else
                value = compare_planar_to_indexed(&org,&tile[i],2,2);
#endif
                if(value < best_value){
                    best_value = value;
                    best_index = i;
                    best_tile = &tile[i];
                }
            }
            paste_tile(vram, best_tile, color_combination[best_index*2+1], color_combination[best_index*2+0], tx*8, ty*tileheight);
#if 1
            nil_paste_indexed_to_planar(ni, best_tile, tx*8, ty*tileheight);
#elif 1
            nil_paste_planar(ni, best_tile, tx*8, ty*tileheight);
#elif 1
            nil_paste_planar(ni, &tile[25], tx*8, ty*tileheight);
#else
            nil_paste_planar(ni, &org, tx*8, ty*tileheight);
#endif
        }
    }
}


#if 0
static int write_vram(nil_spectrum_vram_t *vram, int tileheight, const char *filename){
    FILE *f = fopen(filename,"w");
    fwrite(vram->bitmap, 6144, 1, f);
    fwrite(vram->attributes, 6144/tileheight, 1, f);
    fclose(f);
    return 0;
}
#endif


static int on_option(int opt, const char *arg){
    switch(opt){
    DITHER_PARSE
    case 't':
        opt_tileheight = atoi(arg);
        switch(opt_tileheight){
        case 1:
        case 2:
        case 4:
        case 8:
            return 0;
        default:
            return -1;
        }
    default:
        return -1;
    }
}


int main(int argc,char **argv){
    static const convert_option_t options[] = {
        DITHER_OPTIONS
        {"tileheight",'t',"HEIGHT","Set tileheight to HEIGHT, either 1, 2, 4, or 8 (default is 8)"},
        {NULL}
    };
    convert_t conv;
    nile_t src;
    nile_t dst;
    nil_spectrum_vram_t vram;

    convert_init(&conv, argc, argv, "png", "png", NULL, options, on_option);
    nil_png_read(&src,conv.in);

    nil_resize(&dst,&src,SPECTRUM_WIDTH,SPECTRUM_HEIGHT);
    quant_image(&dst,&vram,opt_tileheight);

    //nile_drop_planar(&dst);
    nil_png_write(&dst,conv.out);
    //write_vram(&vram,opt_tileheight,"out2.scr");
    return 0;
}
