/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "convert.h"


static void flags_append(char *flags, int *flags_len, const char *s, int len){
    if(len<0)
        len = strlen(s);
    memcpy(flags + *flags_len, s, len);
    *flags_len += len;
    flags[*flags_len] = '\0';
}


static void flags_append_char(char *flags, int *flags_len, int ch){
    flags[(*flags_len)++] = ch;
    flags[*flags_len] = '\0';
}


static void usage(convert_t *c, int error){
    FILE *f = error ? stderr : stdout;
    int i;
    fprintf(f, "usage: %s [optional arguments] [<input.%s>] [<output.%s>]\n", c->progname, c->inext, c->outext);
    fprintf(f, "\n"
               "Convert a %s image info a %s image.\n", c->inext, c->outname);
    fprintf(f, "\n"
               "positional arguments:\n"
               "  <input.%s>                  input file (use - or leave empty for stdin)\n"
               "  <output.%s>                 output file (use - or leave empty for stdout)\n", c->inext, c->outext);
    fprintf(f, "\n"
               "optional arguments:\n"
               "  -h --help                    show this help message and exit\n"
               "  -V --version                 show program's version number (" VERSION ") and exit\n");
    if(c->opts)
        for(i=0;c->opts[i].name;++i){
            char buffer[256];
            if(c->opts[i].arg){
                sprintf(buffer, "-%c --%s <%s>",c->opts[i].val,c->opts[i].name,c->opts[i].arg);
            }else{
                sprintf(buffer, "-%c --%s",c->opts[i].val,c->opts[i].name);
            }
            fprintf(f, "  %-28s %s\n",buffer,c->opts[i].description);
        }
}


void convert_init(convert_t *conv,
                  int argc,
                  char **argv,
                  const char *inext,
                  const char *outext,
                  const char *outname,
                  const convert_option_t *opts,
                  convert_option_func_t optfunc)
{
    struct option ent[256];
    int ent_count;
    char flags[4096];
    int flags_len = 0;
    int i;

    conv->progname = argv[0];
    conv->inext = inext;
    conv->outext = outext;
    conv->outname = outname ? outname : outext;
    conv->opts = opts;

    ent[0].name = "help";
    ent[0].has_arg = no_argument;
    ent[0].flag = NULL;
    ent[0].val = 'h';
    ent[1].name = "version";
    ent[1].has_arg = no_argument;
    ent[1].flag = NULL;
    ent[1].val = 'V';
    flags_append(flags, &flags_len, "hV", 2);
    ent_count = 2;

    if(opts)
        for(i=0;opts[i].name;++i){
            ent[ent_count].name = opts[i].name;
            ent[ent_count].has_arg = opts[i].arg ? required_argument : no_argument;
            ent[ent_count].flag = NULL;
            ent[ent_count].val = opts[i].val;
            ++ent_count;
            flags_append_char(flags, &flags_len, opts[i].val);
            if(opts[i].arg)
                flags_append_char(flags, &flags_len, ':');
        }

    ent[ent_count].name = NULL;
    for(;;){
        int c = getopt_long(argc, argv, flags, ent, NULL);
        if(c<0)
            break;
        switch(c){
        case 'h':
            usage(conv, 0);
            exit(0);
        case 'V':
            printf(VERSION "\n");
            exit(0);
        default:
            if(!optfunc || optfunc(c,optarg)<0){
                usage(conv, 0);
                exit(1);
            }
            break;
        }
    }
    if(optind+0 < argc && strcmp(argv[optind+0],"-")){
        conv->in = fopen(argv[optind+0],"rb");
    }else{
        conv->in = stdin;
    }
    if(optind+1 < argc && strcmp(argv[optind+1],"-")){
        conv->out = fopen(argv[optind+1],"wb");
    }else{
        conv->out = stdout;
    }
}


void convert_done(convert_t *conv){
    fclose(conv->in);
    fclose(conv->out);
}


void convert_usage(convert_t *conv){
    usage(conv, 0);
}
