/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <unistd.h>
#include <stdlib.h>

#include "debug.h"
#include "heap.h"

#define MINSIZE 16

heap_t *heap_init(heap_t *h, heap_type_t type, heap_compare_func_t compare)
{
    h->size = MINSIZE;
    h->count = 0;
    h->data = malloc(h->size * sizeof(void*));
    h->type = type;
    h->compare = compare;
    return h;
}


void heap_done_full(heap_t *h, heap_notify_func_t notify){
    if(notify){
#if 0
        while(h->count>0)
            notify(heap_pop(h));
#else
        int i;
        for(i=0;i<h->count;++i)
            notify(h->data[i]);
#endif
    }
    free(h->data);
}


void heap_add(heap_t *h, void *value){
    off_t index;
    off_t parent;

    /* Grow the heap if it's too small to hold all the data. */
    if(h->count >= h->size){
        h->size <<= 1;
        h->data = realloc(h->data, sizeof(void*) * h->size);
        //DEBUG_p("grow to %d",(int)h->size);
    }

    /* Find out where to put the element and put it. */
    for(index=h->count++;index>0;index=parent){
        parent = (index - 1) >> 1;
        if(HEAP_MIN == h->type){
            if(h->compare(h->data[parent], value) <= 0)
                break;
        }else{
            if(h->compare(h->data[parent], value) >= 0)
                break;
        }
        h->data[index] = h->data[parent];
    }
    h->data[index] = value;
}

/** Removes and returns the top item from the heap. */
void *heap_pop(heap_t *h){
    void *value;
    off_t index;
    off_t swap;
    off_t other;

#ifdef _USE_DEBUG
    if(h->count < 100)
        DEBUG_p("count is %d",(int)h->count);
#endif

    if(h->count <= 0)
        return NULL;

    value = heap_peak(h);

    /* Remove the last element. */
    void *temp = h->data[--h->count];

    /* Shrink the heap if it's consuming too much memory. */
    if ((h->count <= (h->size >> 2)) && (h->size > MINSIZE)){
        h->size >>= 1;
        h->data = realloc(h->data, sizeof(void*) * h->size);
        //DEBUG_p("shrink to %d",(int)h->size);
    }

    /* Reorder the elements. */
    for(index=0;;index=swap){
        /* Find the child to swap with. */
        swap = (index << 1) + 1;
        /* If there are no children, the heap is reordered. */
        if(swap >= h->count)
            break;
        other = swap + 1;
        if(HEAP_MIN == h->type){
            if ((other < h->count) && h->compare(h->data[other], h->data[swap]) < 0)
                swap = other;
            if(h->compare(temp,h->data[swap]) <= 0)
                break;
        }else{
            if ((other < h->count) && h->compare(h->data[other], h->data[swap]) > 0)
                swap = other;
            if(h->compare(temp,h->data[swap]) >= 0)
                break;
        }

        h->data[index] = h->data[swap];
    }
    h->data[index] = temp;
    return value;
}
