/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <float.h>
#include <math.h>
#include <string.h>                     /* memcpy() */

#include "common.h"
#include "debug.h"
#include "ni-dual.h"
#include "nil-dither.h"
#include "nil-png-w.h"


typedef struct{
    int index1;
    int index2;
    uint8_t value[4];
}dual_color_t;


static void
_color_blend(uint8_t *dst, const uint8_t *src1, const uint8_t *src2)
{
    int r = (src1[0] + src2[0]) / 2;
    int g = (src1[1] + src2[1]) / 2;
    int b = (src1[2] + src2[2]) / 2;
    dst[0] = r;
    dst[1] = g;
    dst[2] = b;
}


/*
static double color_distance(const uint8_t *src1, const uint8_t *src2){
    int r = src1[0] - src2[0];
    int g = src1[1] - src2[1];
    int b = src1[2] - src2[2];
    return sqrt(r*r + g*g + b*b);
}
*/


/*
static int get_best_color(const dual_color_t *dcol, int dlen, const uint8_t *value){
    double best_distance = DBL_MAX;
    int best_index = -1;
    int i;
    for(i=0;i<dlen;++i){
        double distance = color_distance(value,dcol[i].value);
        if(distance < best_distance){
            best_distance = distance;
            best_index = i;
        }
    }
    return best_index;
}
*/


static int
_get_exact_color(const dual_color_t *dcol, int dlen, const uint8_t *value){
    int i;
    for(i=0;i<dlen;++i)
        if(!memcmp(dcol[i].value,value,3))
            return i;
    return -1;
}


static void
_setup_mixed_palette(dual_color_t *dcol, nile_t *dst, nile_t *src1, nile_t *src2)
{
    int dlen = 0;
    int i;
    for(i=0;i<src1->colors;++i){
        int j;
        for(j=0;j<src2->colors;++j){
            _color_blend(dcol[dlen].value, src1->palette + i*src1->bpc, src2->palette + j*src2->bpc);
            if(_get_exact_color(dcol, dlen, dcol[dlen].value)<0){
                dcol[dlen].index1 = i;
                dcol[dlen].index2 = j;
                ++dlen;
            }
        }
    }
    DEBUG_p("mixed %d unique colors",dlen);
    nile_ensure_palette(dst, dlen, 0);
    for(i=0;i<dlen;++i)
        memcpy(dst->palette + i*dst->bpc, dcol[i].value, dst->bpc);
}


int ni_encode_dual(nile_t *dst1,nile_t *dst2,const nile_t *src, nil_dither_type_t opt_dither){
    nile_t tmp;
    dual_color_t dcol[256*256];
    int y;

    nile_init(&tmp);
    tmp.width = src->width;
    tmp.height = src->height;
    tmp.format = src->format;
    tmp.bpc = src->bpc;
    _setup_mixed_palette(dcol, &tmp, dst1, dst2);
    tmp.planar = src->planar;
    //nil_png_save(&tmp,"tmp-in.png");
    dither(&tmp, opt_dither, 0, 0, 4);
    tmp.planar = NULL;
    //nil_png_save(&tmp,"tmp-out.png");

    {
        const uint8_t *p0 = tmp.indexed;
        int offs1 = 0;
        int offs2 = 0;
        int stride0 = nile_get_indexed_stride(&tmp);
        int stride1 = nile_get_indexed_stride(dst1);
        int stride2 = nile_get_indexed_stride(dst2);
        for(y=0;y<tmp.height;++y){
            int x;
            for(x=0;x<tmp.width;++x){
                int idx = p0[x];
                if((x+y)&1){
                    dst1->indexed[offs1+x] = dcol[idx].index2;
                    dst2->indexed[offs2+x] = dcol[idx].index1;
                }else{
                    dst1->indexed[offs1+x] = dcol[idx].index1;
                    dst2->indexed[offs2+x] = dcol[idx].index2;
                }
            }
            p0 += stride0;
            offs1 += stride1;
            offs2 += stride2;
        }
    }

    nile_done(&tmp);
    return 0;
}
