/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <math.h>

#include "debug.h"
#include "nil-blur.h"

void nil_blur_gaussian_matrix(double *matrix, int size, double sigma){
    int radius = (size-1)/2;
    //double s = 2.0 * sigma * sigma;
    double s = sigma * sigma * radius;
    double sum = 0;
    int x;

    for(x = -radius; x <= radius; ++x) {
        int y;
        for (y = -radius; y <= radius; ++y) {
            double r = sqrt(x * x + y * y);
            sum += matrix[x+radius + (y+radius)*size] = (exp(-(r * r) / s)) / (M_PI * s);
        }
    }

    for (x=0; x< size*size; ++x)
        matrix[x] /= sum;
}


static const uint8_t *_get_pixel_indexed(nile_t *ni, int x, int y, int stride){
    return ni->palette + ni->indexed[x + y*stride]*ni->bpc;
}


void nil_blur_apply_matrix(nile_t *ni,double *matrix,int size){
    nile_t tmp;
    int src_stride = nile_get_planar_stride(ni);
    uint8_t *dst;
    int dst_skip;
    int radius = (size-1)/2;
    int py;

    nile_init(&tmp);
    nile_ensure_planar(&tmp, ni->width, ni->height, ni->bpc);
    src_stride = nile_get_planar_stride(ni);

    dst = tmp.planar;
    tmp.planar = ni->planar;
    ni->planar = dst;
    ni->planar_stride = tmp.planar_stride;
    dst_skip = nile_get_planar_skip(ni);

    for(py=0;py<ni->height;++py){
        int px;
        for(px=0;px<ni->width;++px){
            double channel[4];
            int my;
            int i;
            for(i=0;i<ni->bpc;++i)
                channel[i] = 0;
            for(my=-radius;my<=radius;++my){
                int y = py+my;
                int mx;
                if(y<0){
                    y = 0;
                }else if(y>=ni->height){
                    y = ni->height-1;
                }
                for(mx=-radius;mx<=radius;++mx){
                    int x = px+mx;
                    const uint8_t *src;
                    double val;
                    if(x<0){
                        x = 0;
                    }else if(x>=ni->width){
                        x = ni->width-1;
                    }
                    src = &tmp.planar[x*tmp.bpc + y*src_stride];
                    val = matrix[mx+radius + (my+radius)*size];
                    for(i=0;i<ni->bpc;++i)
                        channel[i] += val * src[i];
                }
            }
            for(i=0;i<ni->bpc;++i)
                dst[i] = channel[i];
            dst += ni->bpc;
        }
        dst += dst_skip;
    }
    nile_done(&tmp);
}


/*
  cycles
  9 323 102 208
  8 359 723 008
  3 672 281 088

*/
void nil_blur_apply_matrix_indexed(nile_t *ni,double *matrix,int size){
    int src_stride = nile_get_indexed_stride(ni);
    uint8_t *dst = ni->planar;
    int dst_skip = nile_get_planar_skip(ni);
    int radius = (size-1)/2;
    int py;
    for(py=0;py<ni->height;++py){
        int px;
        for(px=0;px<ni->width;++px){
#if 1
            double channel[4];
            int my;
            int i;
            for(i=0;i<ni->bpc;++i)
                channel[i] = 0;
            for(my=-radius;my<=radius;++my){
                int y = py+my;
                int mx;
                if(y<0){
                    y = 0;
                }else if(y>=ni->height){
                    y = ni->height-1;
                }
                for(mx=-radius;mx<=radius;++mx){
                    int x = px+mx;
                    const uint8_t *src;
                    double val;
                    if(x<0){
                        x = 0;
                    }else if(x>=ni->width){
                        x = ni->width-1;
                    }
                    src = _get_pixel_indexed(ni, x, y, src_stride);
                    val = matrix[mx+radius + (my+radius)*size];
                    for(i=0;i<ni->bpc;++i)
                        channel[i] += val * src[i];
                }
            }
            for(i=0;i<ni->bpc;++i)
                dst[i] = channel[i];
#else
            const uint8_t *src = _get_pixel_indexed(ni, px, py, src_stride);
            int i;
            for(i=0;i<ni->bpc;++i)
                dst[i] = src[i];
#endif
            dst += ni->bpc;
        }
        dst += dst_skip;
    }
}


void nil_blur_gaussian(nile_t *ni,double sigma){
    double matrix[5*5];
    nil_blur_gaussian_matrix(matrix,5,sigma);
    nil_blur_apply_matrix(ni, matrix, 5);
}


void nil_blur_gaussian_indexed(nile_t *ni,double sigma){
    double matrix[5*5];
    nil_blur_gaussian_matrix(matrix,5,sigma);
    nil_blur_apply_matrix_indexed(ni, matrix, 5);
}
