/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <float.h>                      /* DBL_MAX */
#include <math.h>
#include <stdlib.h>                     /* calloc(), free() */

#include "nil-dither.h"

#ifndef MAX
 #define MAX(a,b) ((a)>(b)?(a):(b))
#endif
#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#define MAXDIFF 112

static double _rgb_diff(uint8_t *rgb1, uint8_t *rgb2, int bpc){
    double diff = 0;
    int i;
    for(i=0;i<bpc;++i){
        double d = rgb1[i]-rgb2[i];
        diff += d*d;
    }
    return diff;
}


#if 0
static double _rgb_int_diff(int *rgb1, uint8_t *rgb2, int bpc){
    double diff = 0;
    int i;
    for(i=0;i<bpc;++i){
        double d = rgb1[i]-rgb2[i];
        diff += d*d;
    }
    return diff;
}
#endif


static double _rgb_double_diff(double *rgb1, uint8_t *rgb2, int bpc){
    double diff = 0;
    int i;
    for(i=0;i<bpc;++i){
        double d = rgb1[i]-rgb2[i];
        diff += d*d;
    }
    return diff;
}


static int _rgb_nearest(uint8_t *rgb, int bpc, uint8_t *palette, int colors) {
    double best_diff = DBL_MAX;
    int nearest = 0;
    int i;
    for(i=0;i<colors;++i){
        double diff = _rgb_diff(rgb, palette + i*bpc, bpc);
        if(diff < best_diff){
            best_diff = diff;
            nearest = i;
        }
    }
    return nearest;
}


#if 0
static int _rgb_int_nearest(int *rgb, int bpc, uint8_t *palette, int colors) {
    double best_diff = DBL_MAX;
    int nearest;
    int i;
    for(i=0;i<colors;++i){
        double diff = _rgb_int_diff(rgb, palette + i*bpc, bpc);
        if(diff < best_diff){
            best_diff = diff;
            nearest = i;
        }
    }
    return nearest;
}
#endif


static int _rgb_double_nearest(double *rgb, int bpc, uint8_t *palette, int colors) {
    double best_diff = DBL_MAX;
    int nearest = 0;
    int i;
    for(i=0;i<colors;++i){
        double diff = _rgb_double_diff(rgb, palette + i*bpc, bpc);
        if(diff < best_diff){
            best_diff = diff;
            nearest = i;
        }
    }
    return nearest;
}


int nil_dither_none(nile_t *ni){
    uint8_t *src = ni->planar;
    int src_skip = nile_get_planar_skip(ni);
    uint8_t *dst;
    int dst_skip;
    int y;

    nile_ensure_indexed(ni,0,0,0);
    dst = ni->indexed;
    dst_skip = nile_get_indexed_stride(ni) - ni->width;

    for(y=ni->height;y>0;--y){
        int x;
        for(x=ni->width;x>0;--x){
            *dst++ = _rgb_nearest(src, ni->bpc, ni->palette, ni->colors);
            src += ni->bpc;
        }
        src += src_skip;
        dst += dst_skip;
    }
    return 0;
}


int nil_dither_ordered4x4(nile_t *ni, int xoffs, int yoffs, int channel_depth){
    static const int bayer[4*4] = {
         1,  9,  3, 11,
        13,  5, 15,  7,
         4, 12,  2, 10,
        16, 8,  14,  6
    };
    uint8_t *src = ni->planar;
    int src_skip = nile_get_planar_skip(ni);
    uint8_t *dst;
    int dst_skip;
    int y;

    nile_ensure_indexed(ni,0,0,0);
    dst = ni->indexed;
    dst_skip = nile_get_indexed_stride(ni) - ni->width;

    for(y=ni->height;y>0;--y){
        int x;
        for(x=ni->width;x>0;--x){
            int d = 3 * bayer[((y+yoffs)%4)*4 + (x+xoffs)%4];
            uint8_t channel[4];
            int i;
            for(i=0;i<ni->bpc;++i)
                channel[i] = MIN(d + src[i],255);

            src += ni->bpc;
            *dst++ = _rgb_nearest(channel, ni->bpc, ni->palette, ni->colors);
        }
        src += src_skip;
        dst += dst_skip;
    }
    return 0;
}


/*
  cycles
  1 015 177 764
*/
int nil_dither_ordered8x8(nile_t *ni, int xoffs, int yoffs, int channel_depth){
    static const int bayer[8*8] = {
         1, 49, 13, 61,  4, 52, 16, 64,
        33, 17, 45, 29, 36, 20, 48, 32,
         9, 57,  5, 53, 12, 60,  8, 56,
        41, 25, 37, 21, 44, 28, 40, 24,
         3, 51, 15, 63,  2, 50, 14, 62,
        35, 19, 47, 31, 34, 18, 46, 30,
        11, 59,  7, 55, 10, 58,  6, 54,
        43, 27, 39, 23, 42, 26, 38, 22
    };
    uint8_t *src = ni->planar;
    int src_skip = nile_get_planar_skip(ni);
    uint8_t *dst;
    int dst_skip;
    int y;

    nile_ensure_indexed(ni,0,0,0);
    dst = ni->indexed;
    dst_skip = nile_get_indexed_skip(ni);

    for(y=ni->height;y>0;--y){
        int x;
        for(x=ni->width;x>0;--x){
            int d = bayer[((y+yoffs)%8)*8 + (x+xoffs)%8];
            uint8_t channel[4];
            int i;
            for(i=0;i<ni->bpc;++i)
                channel[i] = MIN(d + src[i],255);

            src += ni->bpc;
            *dst++ = _rgb_nearest(channel, ni->bpc, ni->palette, ni->colors);
        }
        src += src_skip;
        dst += dst_skip;
    }
    return 0;
}


static void _add_scaled(double *dst, double *src, int bpc, double scale){
    int i;
    for(i=0;i<bpc;++i)
        dst[i] += scale * src[i];
}


static void _apply_error_diffuse_floyd_steinberg(double *diffmap, double *diff, int x, int y, int width, int height, int bpc){
    int stride = width * bpc;
    if(x+1 < width)
        _add_scaled(&diffmap[bpc], diff, bpc, 7.0/16);
    if(y+1 < height){
        if(x-1 >= 0)
            _add_scaled(&diffmap[stride - bpc], diff, bpc, 3.0/16);
        _add_scaled(&diffmap[stride], diff, bpc, 5.0/16);
        if(x+1 < width)
            _add_scaled(&diffmap[stride + bpc], diff, bpc, 1.0/16);
    }
}


static void _apply_error_diffuse_jarvis_judice_ninke(double *diffmap, double *diff, int x, int y, int width, int height, int bpc){
    int stride = width * bpc;
    if(x+1 < width){
        _add_scaled(&diffmap[bpc], diff, bpc, 7.0/48);
        if(x+2 < width)
            _add_scaled(&diffmap[bpc], diff, bpc, 5.0/48);
    }
    if(y+1 < height){
        if(x-1 >= 0){
            if(x-2 >= 0)
                _add_scaled(&diffmap[stride - bpc*2], diff, bpc, 3.0/48);
            _add_scaled(&diffmap[stride - bpc], diff, bpc, 5.0/48);
        }
        _add_scaled(&diffmap[stride], diff, bpc, 7.0/48);
        if(x+1 < width){
            _add_scaled(&diffmap[stride + bpc], diff, bpc, 5.0/48);
            if(x+2 < width)
                _add_scaled(&diffmap[stride + bpc*2], diff, bpc, 3.0/48);
        }
        if(y+2 < height){
            if(x-1 >= 0){
                if(x-2 >= 0)
                    _add_scaled(&diffmap[stride*2 - bpc*2], diff, bpc, 1.0/48);
                _add_scaled(&diffmap[stride*2 - bpc], diff, bpc, 3.0/48);
            }
            _add_scaled(&diffmap[stride*2], diff, bpc, 5.0/48);
            if(x+1 < width){
                _add_scaled(&diffmap[stride*2 + bpc], diff, bpc, 3.0/48);
                if(x+2 < width)
                    _add_scaled(&diffmap[stride*2 + bpc*2], diff, bpc, 1.0/48);
            }
        }
    }
}


static void _apply_error_diffuse_stucki(double *diffmap, double *diff, int x, int y, int width, int height, int bpc){
    int stride = width * bpc;
    if(x+1 < width){
        _add_scaled(&diffmap[bpc], diff, bpc, 8.0/42);
        if(x+2 < width)
            _add_scaled(&diffmap[bpc], diff, bpc, 4.0/42);
    }
    if(y+1 < height){
        if(x-1 >= 0){
            if(x-2 >= 0)
                _add_scaled(&diffmap[stride - bpc*2], diff, bpc, 2.0/42);
            _add_scaled(&diffmap[stride - bpc], diff, bpc, 4.0/42);
        }
        _add_scaled(&diffmap[stride], diff, bpc, 8.0/42);
        if(x+1 < width){
            _add_scaled(&diffmap[stride + bpc], diff, bpc, 4.0/42);
            if(x+2 < width)
                _add_scaled(&diffmap[stride + bpc*2], diff, bpc, 2.0/42);
        }
        if(y+2 < height){
            if(x-1 >= 0){
                if(x-2 >= 0)
                    _add_scaled(&diffmap[stride*2 - bpc*2], diff, bpc, 1.0/42);
                _add_scaled(&diffmap[stride*2 - bpc], diff, bpc, 2.0/42);
            }
            _add_scaled(&diffmap[stride*2], diff, bpc, 4.0/42);
            if(x+1 < width){
                _add_scaled(&diffmap[stride*2 + bpc], diff, bpc, 2.0/42);
                if(x+2 < width)
                    _add_scaled(&diffmap[stride*2 + bpc*2], diff, bpc, 1.0/42);
            }
        }
    }
}


typedef void(*apply_error_diffuse_func_t)(double*,double*,int,int,int,int,int);


static void _dither_error_diffuse(nile_t *ni, apply_error_diffuse_func_t apply_error_diffuse){
    uint8_t *src = ni->planar;
    int src_skip = nile_get_planar_skip(ni);
    uint8_t *dst;
    int dst_skip;
    double *diffmap_mem;
    double *diffmap;
    int y;

    nile_ensure_indexed(ni,0,0,0);
    dst = ni->indexed;
    dst_skip = nile_get_indexed_skip(ni);

    diffmap = diffmap_mem = (double*)calloc(ni->width * ni->height * ni->bpc, sizeof(double));

    for(y=0;y<ni->height;++y){
        int x;
        for(x=0;x<ni->width;++x){
            double channel[4];
            int idx;
            int i;
            diffmap = &diffmap_mem[(x + y*ni->width) * ni->bpc];
            for(i=0;i<ni->bpc;++i){
                double val = round(diffmap[i]);
                channel[i] = src[i] + MIN(MAX(val,-MAXDIFF),MAXDIFF);
            }
            src += ni->bpc;
            idx = _rgb_double_nearest(channel, ni->bpc, ni->palette, ni->colors);
            *dst++ = idx;
            for(i=0;i<ni->bpc;++i)
                channel[i] -= ni->palette[idx*ni->bpc + i];
            apply_error_diffuse(diffmap, channel, x, y, ni->width, ni->height, ni->bpc);
        }
        src += src_skip;
        dst += dst_skip;
    }
    free(diffmap_mem);
}


void nil_dither_floyd_steinberg(nile_t * ni){
    _dither_error_diffuse(ni, _apply_error_diffuse_floyd_steinberg);
}


void nil_dither_jarvis_judice_ninke(nile_t *ni){
    _dither_error_diffuse(ni, _apply_error_diffuse_jarvis_judice_ninke);
}


void nil_dither_stucki(nile_t *ni){
    _dither_error_diffuse(ni, _apply_error_diffuse_stucki);
}
