/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <stdio.h>
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memset() */
#include <png.h>

#include "nil-png-w.h"


static int
_convert_colors(png_color *rgb, uint8_t *alpha, const uint8_t *src, nile_format_t src_format, int src_bpc, int count)
{
    int o_r;
    int o_g;
    int o_b;
    int o_a;
    int i;
    if(src_bpc > 4)
        return -1;
    o_r = NILE_OFFSET_red(src_format);
    o_g = NILE_OFFSET_green(src_format);
    o_b = NILE_OFFSET_blue(src_format);
    o_a = NILE_OFFSET_alpha(src_format);
    for(i=0;i<count;++i){
        rgb[i].red = src[o_r];
        rgb[i].green = src[o_g];
        rgb[i].blue = src[o_b];
        if(o_a != NILE_CHANNEL_MISSING && alpha){
            alpha[i] = src[o_a];
        }else if(alpha){
            alpha[i] = 255;
        }
        src += src_bpc;
    }
    return 0;
}


int
_write_indexed(nile_t *ni,png_structp png,png_infop info){
    png_color png_palette[256];
    uint8_t png_alpha[256];

    png_set_IHDR(png,info,
                 ni->width,
                 ni->height,
                 8,PNG_COLOR_TYPE_PALETTE,
                 PNG_INTERLACE_NONE,
                 PNG_COMPRESSION_TYPE_DEFAULT,
                 PNG_FILTER_TYPE_DEFAULT);

    _convert_colors(png_palette,png_alpha,ni->palette,ni->format,ni->bpc,ni->colors);
    png_set_PLTE(png,info,png_palette,ni->colors);
    if(NILE_HAS_ALPHA(ni->format))
        png_set_tRNS(png,info,png_alpha,ni->colors,NULL);

    png_write_info(png,info);
    {
        png_byte** rows = malloc(ni->height*sizeof(png_byte*));
        int stride = nile_get_indexed_stride(ni);
        int i;
        for(i=0;i<ni->height;++i)
            rows[i] = ni->indexed + i*stride;
        png_write_rows(png,rows,ni->height);
        free(rows);
    }
    return 0;
}


int
_write_planar(nile_t *ni,png_structp png,png_infop info){
    const uint8_t *src = ni->planar;
    int src_stride;
    png_color** png_rows;
    int png_stride;
    int i;
    if(!src)
        return -1;
    src_stride = nile_get_planar_stride(ni);
    png_rows = malloc(ni->height*sizeof(png_color*));
    png_stride = ni->width * sizeof(png_color);

    png_set_IHDR(png,info,
                 ni->width,
                 ni->height,
                 8,PNG_COLOR_TYPE_RGB,
                 PNG_INTERLACE_NONE,
                 PNG_COMPRESSION_TYPE_DEFAULT,
                 PNG_FILTER_TYPE_DEFAULT);
    png_write_info(png,info);
    for(i=0;i<ni->height;++i){
        png_rows[i] = i ? png_rows[i-1] + png_stride : malloc(png_stride * ni->height * sizeof(png_color));
        _convert_colors(png_rows[i],NULL,src,ni->format,ni->bpc,ni->width);
        src += src_stride;
    }
    png_write_rows(png,(png_bytepp)png_rows,ni->height);
    free(png_rows[0]);
    free(png_rows);
    return 0;
}


int
nil_png_write(nile_t *ni, FILE *f)
{
    png_structp png;
    png_infop info;
    int res;

    png = png_create_write_struct(PNG_LIBPNG_VER_STRING,NULL,NULL,NULL);
    if(!png)
        return -1;
    info = png_create_info_struct(png);
    if(!info){
        png_destroy_write_struct(&png,(png_infopp)NULL);
        return -1;
    }
    png_init_io(png,f);

    if(ni->palette && ni->indexed){
        res = _write_indexed(ni,png,info);
    }else{
        res = _write_planar(ni,png,info);
    }

    if(0==res)
        png_write_end(png,info);
    png_destroy_write_struct(&png,&info);
    return res;
}


int
nil_png_save(nile_t *ni, const char *filename)
{
    FILE* f = fopen(filename,"wb");
    int res;
    if(!f)
        return -1;
    res = nil_png_write(ni, f);
    fclose(f);
    return res;
}
