/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <float.h>                      /* DBL_MAX */
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memcmp() */

#include "debug.h"
#include "heap.h"
#include "nil-quant.h"
#include "pool.h"


#define NIL_QUANT_TEST_POOL


static void
_copy_color_shift(uint8_t *dst, const uint8_t *src, int bpc, int shifts){
    int i;
    if(shifts > 0){
        for(i=0;i<bpc;++i){
            int b = src[i] << shifts;
            int bits = 8-shifts;
            while(bits < 8){
                b |= b>>bits;
                bits <<= 1;
            }
            dst[i] = b;
        }
    }else if(shifts < 0){
        for(i=0;i<bpc;++i)
            dst[i] = src[i] >> -shifts;
    }else{
        memcpy(dst, src, bpc);
    }
}


static void _color_mix_weighted(uint8_t *dst, uint8_t *src1, int weight1, uint8_t *src2, int weight2, int bpc){
    int i;
    if(weight1 || weight2)
        for(i=0;i<bpc;++i)
            dst[i] = (src1[i]*weight1 + src2[i]*weight2) / (weight1+weight2);
}


static int color3_equal(const void *color1, const void *color2){
    return !memcmp(color1, color2, 3);
}

static unsigned int color3_hash(const void* data){
    const uint8_t *color = data;
    unsigned int h = color[0];
    h = (h<<5)-h + color[1];
    return (h<<5)-h + color[2];
}


typedef struct _color_node_t color_node_t;
typedef struct _color_pair_t color_pair_t;

struct _color_node_t{
    int refcount;
    uint8_t color[NILE_MAX_BYTES_PER_COLOR];
    int count;                          /**< Times used, a negative number means the color is mandatory. */
    int valid;
};


static color_node_t *color_node_new(const uint8_t *color, int bpc, int count){
    color_node_t *node = (color_node_t*)malloc(sizeof(color_node_t));
    node->refcount = 1;
    memcpy(node->color, color, bpc);
    node->count = count;
    node->valid = 1;
    return node;
}


static color_node_t *color_node_ref(color_node_t *node){
    ++node->refcount;
    return node;
}


static void color_node_unref(color_node_t *node){
    if(0 == --node->refcount){
        //DEBUG_p("dropping color");
        free(node);
    }
}


static void color_node_add(color_node_t *node, int count){
    if(node->count < 0)
        return;
    if(count < 0){
        node->count = -1;
    }else{
        node->count += count;
    }
}


static double color_node_score(const color_node_t *node){
    if(!node->valid)
        return -1;
    if(node->count < 0)
        return DBL_MAX;
    return node->count;
}


static int color_node_compare(const void *data1, const void *data2){
    double score1 = color_node_score(data1);
    double score2 = color_node_score(data2);
    if(score1 == score2){
        return 0;
    }else if(score1 > score2){
        //DEBUG_p("%f > %f",score1,score2);
        return 1;
    }else{
        //DEBUG_p("%f < %f",score1,score2);
        return -1;
    }
}



struct _color_pair_t{
    color_node_t *color[2];
    double score;
#ifdef NIL_QUANT_TEST_POOL
    pool_t *pool;
#endif
};


static double color_pair_score(const color_pair_t *pair, int total_pixels){
    double distance;
    int usage;
    if(0 == pair->color[0]->count && 0 == pair->color[1]->count)
        return 0;
    if(pair->color[0]->count < 0 && pair->color[1]->count < 0)
        return DBL_MAX;
    distance = nile_color_distance(pair->color[0]->color, NILE_FORMAT_r8g8b8, pair->color[1]->color, NILE_FORMAT_r8g8b8);
    usage = pair->color[0]->count + pair->color[1]->count;
    return distance + 80*(double)usage/total_pixels;
}


static color_pair_t *
#ifdef NIL_QUANT_TEST_POOL
color_pair_new(pool_t *pool, color_node_t *node1, color_node_t *node2, int bpc, int total_pixels)
#else
color_pair_new(color_node_t *node1, color_node_t *node2, int bpc, int total_pixels)
#endif
{
#ifdef NIL_QUANT_TEST_POOL
    color_pair_t *pair = pool_lease(pool);
    pair->pool = pool;
#else
    color_pair_t *pair = (color_pair_t*)malloc(sizeof(color_pair_t));
#endif
    pair->color[0] = node1;
    pair->color[1] = node2;
    pair->score = color_pair_score(pair, total_pixels);
    return pair;
}


static void color_pair_free(color_pair_t *pair){
    color_node_unref(pair->color[0]);
    color_node_unref(pair->color[1]);
#ifdef NIL_QUANT_TEST_POOL
    pool_return(pair->pool, pair);
#else
    free(pair);
#endif
}


static int color_pair_compare(const void *data1, const void *data2){
    const color_pair_t *pair1 = data1;
    const color_pair_t *pair2 = data2;
    if(pair1->score == pair2->score){
        return 0;
    }else if(pair1->score > pair2->score){
        return 1;
    }else{
        return -1;
    }
}


typedef int(*child_from_color_func_t)(int level,const uint8_t *color);
typedef void(*foreach_func_t)(void *userdata, color_node_t *node);


typedef struct{
    const uint8_t *color;
    int bpc;
    double best_value;
    color_node_t *best_node;
}nearest_data_t;

static int _nearest_cb(void *key, void *val, void *userdata){
    nearest_data_t *d = userdata;
    color_node_t *node = val;
    double value = nile_color_variance(d->color, NILE_FORMAT_r8g8b8, node->color, NILE_FORMAT_r8g8b8);
    if(value < d->best_value){
        d->best_value = value;
        d->best_node = node;
    }
    return 0;
}

static color_node_t *
_get_nearest(nil_quant_set_t *set, const uint8_t *color, const nile_t *ni){
    nearest_data_t d;
    d.best_node = hashmap_get(&set->color_map, color);
    if(!d.best_node){
        d.color = color;
        d.bpc = ni->bpc;
        d.best_value = DBL_MAX;
        hashmap_foreach(&set->color_map, _nearest_cb, &d);
    }
    return d.best_node;
}


static int _colors_to_heap_cb(void *key, void *val, void *userdata){
    heap_t *h = userdata;
    color_node_t *node = val;
    //DEBUG_p("adding node with score %f",color_node_score(node));
    heap_add(h, node);
    return 0;
}

static void
_colors_to_heap(nil_quant_set_t *set, heap_t *h){
    hashmap_foreach(&set->color_map, _colors_to_heap_cb, h);
}


nil_quant_set_t *
nil_quant_set_init(nil_quant_set_t *set, int channel_depth){
    set->channel_depth = channel_depth;
    hashmap_init(&set->color_map,color3_hash, color3_equal, NULL, (hashmap_notify_func_t)color_node_unref);
    set->total_pixels = 0;
    return set;
}


void
nil_quant_set_done(nil_quant_set_t *set){
    hashmap_done(&set->color_map);
}


#define combin(n) (((n)*(n)-(n))/2)


typedef struct{
    color_node_t *node;
    heap_t *pair_heap;
    pool_t *pair_pool;
    int bpc;
    int total_pixels;
}add_data_t;

static int _add_cb(void *key, void *value, void *userdata){
    add_data_t *d = userdata;
    color_node_t *node2 = (color_node_t*)value;
    if(node2->valid){
#ifdef NIL_QUANT_TEST_POOL
        color_pair_t *pair = color_pair_new(d->pair_pool, color_node_ref(d->node), color_node_ref(node2), d->bpc, d->total_pixels);
#else
        color_pair_t *pair = color_pair_new(color_node_ref(d->node), color_node_ref(node2), d->bpc);
#endif
        heap_add(d->pair_heap, pair);
    }
    return 0;
}

static void _add_full(nil_quant_set_t *set, heap_t *pair_heap, pool_t *pair_pool, const uint8_t *color_org, int count, int bpc, int shifts){
    uint8_t color[NILE_MAX_BYTES_PER_COLOR];
    add_data_t d;
    _copy_color_shift(color, color_org, bpc, -shifts);
    d.node = hashmap_get(&set->color_map, color);
    set->total_pixels += count;
    if(d.node){
        color_node_add(d.node, count);
        return;
    }
    d.node = color_node_new(color, bpc, count);
    if(pair_heap){
        d.pair_heap = pair_heap;
        d.pair_pool = pair_pool;
        d.bpc = bpc;
        d.total_pixels = set->total_pixels;
        hashmap_foreach(&set->color_map, _add_cb, &d);
    }
    hashmap_add(&set->color_map, d.node->color, d.node);
    //DEBUG_p("colors: %d, pairs: %d (%d)",nil_quant_set_colors(set),(int)(pair_heap?pair_heap->count:-1),combin(nil_quant_set_colors(set)));
}




void nil_quant_set_add(nil_quant_set_t *set,const uint8_t *color, int count, const nile_t *ni){
    _add_full(set, NULL, NULL, color, count, ni->bpc, 8-set->channel_depth);
}


void nil_quant_set_add_pixels(nil_quant_set_t *set, const nile_t *ni){
    uint8_t *src = ni->planar;
    int src_skip = nile_get_planar_skip(ni);
    int y;
    for(y=0;y< ni->height ;++y){
        int x;
        for(x=0;x<ni->width;++x){
            nil_quant_set_add(set, src, 1, ni);
            src += ni->bpc;
        }
        src += src_skip;
    }
}


void nil_quant_set_add_nearest(nil_quant_set_t *set, const uint8_t *color, int count, const nile_t *ni){
    color_node_t *node = _get_nearest(set, color, ni);
    color_node_add(node, count);
}


void nil_quant_set_add_nearest_indexed_pixels(nil_quant_set_t *set, const nile_t *ni){
    uint8_t *src = ni->indexed;
    int src_skip = nile_get_indexed_skip(ni);
    int y;
    for(y=0;y< ni->height ;++y){
        int x;
        for(x=0;x<ni->width;++x)
            nil_quant_set_add_nearest(set, ni->palette + *src++ * ni->bpc, 1, ni);
        src += src_skip;
    }
}


void nil_quant_set_add_nearest_planar_pixels(nil_quant_set_t *set, const nile_t *ni){
    uint8_t *src = ni->planar;
    int src_skip = nile_get_planar_skip(ni);
    int y;
    for(y=0;y< ni->height ;++y){
        int x;
        for(x=0;x<ni->width;++x){
            nil_quant_set_add_nearest(set, src, 1, ni);
            src += ni->bpc;
        }
        src += src_skip;
    }
}


void nil_quant_set_reduce_mix(nil_quant_set_t *set, int colors, const nile_t *ni){
    heap_t color_heap;
    heap_t pair_heap;
#ifdef NIL_QUANT_TEST_POOL
    pool_t pair_pool;
#endif
    if(nil_quant_set_colors(set) <= colors)
        return;
    heap_init(&color_heap, HEAP_MIN, color_node_compare);
    _colors_to_heap(set, &color_heap);
    heap_init(&pair_heap, HEAP_MIN, color_pair_compare);
#ifdef NIL_QUANT_TEST_POOL
    pool_init(&pair_pool, sizeof(color_pair_t), 0);
#endif
    while(color_heap.count > 0){
        color_node_t *node1 = heap_pop(&color_heap);
        int i;
        for(i=0;i<color_heap.count;++i){
            color_node_t *node2 = color_heap.data[i];
#ifdef NIL_QUANT_TEST_POOL
            color_pair_t *pair = color_pair_new(&pair_pool, color_node_ref(node1), color_node_ref(node2), ni->bpc, set->total_pixels);
#else
            color_pair_t *pair = color_pair_new(color_node_ref(node1), color_node_ref(node2), ni->bpc);
#endif
            heap_add(&pair_heap, pair);
        }
    }
    heap_done(&color_heap);
    while(nil_quant_set_colors(set) > colors){
        color_pair_t *pair;
        //DEBUG_p("colors: %d, pairs: %d (%d)",nil_quant_set_colors(set),(int)pair_heap.count,combin(nil_quant_set_colors(set)));
        pair = heap_pop(&pair_heap);
        if(pair->color[0]->valid && pair->color[1]->valid){
            uint8_t color[NILE_MAX_BYTES_PER_COLOR];
            color_node_t *node1 = pair->color[0];
            color_node_t *node2 = pair->color[1];
            int count = node1->count + node2->count;
            DEBUG_p("removing pair with score %f",pair->score);
            set->total_pixels -= node1->count + node2->count;
            node1->valid = 0;
            node2->valid = 0;
            _color_mix_weighted(color, node1->color, node1->count, node2->color, node2->count, ni->bpc);
            hashmap_remove(&set->color_map, node1->color);
            hashmap_remove(&set->color_map, node2->color);
            _add_full(set, &pair_heap, &pair_pool, color, count, ni->bpc, 0);
        }else{
            //DEBUG_p("found invalid pair, pairs left: %d (%d)",(int)pair_heap.count,combin(nil_quant_set_colors(set)));
        }
        color_pair_free(pair);
    }
    heap_done_full(&pair_heap, (heap_notify_func_t)color_pair_free);
    pool_done(&pair_pool);
}


void nil_quant_set_reduce_simple(nil_quant_set_t *set, int colors, const nile_t *ni){
    heap_t heap;
    if(nil_quant_set_colors(set) <= colors){
        //DEBUG_p("only %d colors, no need to reduce",set->colors);
        return;
    }
    heap_init(&heap, HEAP_MIN, color_node_compare);
    _colors_to_heap(set, &heap);
    while(nil_quant_set_colors(set) > colors){
        color_node_t *node = heap_pop(&heap);
        DEBUG_p("colors: %d, removing node with score %f",nil_quant_set_colors(set),color_node_score(node));
        node->valid = 0;
        hashmap_remove(&set->color_map, node->color);
        //--set->colors;
    }
    heap_done_full(&heap, NULL);
}


void nil_quant_set_get_palette(nil_quant_set_t *set, nile_t *ni){
    heap_t heap;
    int i = 0;
    heap_init(&heap, HEAP_MAX, color_node_compare);
    _colors_to_heap(set, &heap);
    nile_ensure_palette(ni, heap.count, 0);
    while(heap.count > 0){
        color_node_t *node = heap_pop(&heap);
        _copy_color_shift(ni->palette + i*ni->bpc, node->color, ni->bpc, 8-set->channel_depth);
        ++i;
    }
    heap_done(&heap);
}


void nil_quant(nile_t *ni, int colors, int channel_depth){
    nil_quant_set_t set;
    DEBUG_p("bytes per color: %d",ni->bpc);
    nil_quant_set_init(&set, channel_depth);
    nil_quant_set_add_pixels(&set, ni);
    DEBUG_p("total colors: %d", nil_quant_set_colors(&set));
    nil_quant_set_reduce_mix(&set, colors, ni);
    nil_quant_set_get_palette(&set, ni);
    nil_quant_set_done(&set);
}
