/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <float.h>                      /* DBL_MAX */
#include <stddef.h>                     /* size_t */
#include <stdlib.h>                     /* malloc(), free() */
#include <string.h>                     /* memcmp(), memcpy() */

#include "nile.h"


#define CHECK_MISMATCH(a,b) (((a)>0 && (b)>0 && (a) != (b)) || ((a)<=0 && (b)<=0))

nile_t *
nile_init(nile_t *ni)
{
    memset(ni,0,sizeof(nile_t));
    return ni;
}


void
nile_done(nile_t *ni)
{
    if(ni->palette)
        free(ni->palette);
    if(ni->bitplanes){
        free(ni->bitplanes[0]);
        free(ni->bitplanes);
    }
    if(ni->indexed)
        free(ni->indexed);
    if(ni->planar)
        free(ni->planar);
}


nile_t *
nile_new()
{
    return (nile_t*)calloc(1,sizeof(nile_t));
}


void
nile_free(nile_t* ni)
{
    nile_done(ni);
    free(ni);
}



int
nile_ensure_palette(nile_t *ni,int colors, int bpc)
{
    if(CHECK_MISMATCH(ni->bpc, bpc))
        return -1;
    if(colors>0)
        ni->colors = colors;
    if(ni->colors <= 0)
        return -1;
    if(bpc>0)
        ni->bpc = bpc;
    if(ni->bpc <= 0)
        return -1;
    if(ni->palette){
        ni->palette = realloc(ni->palette, ni->colors * ni->bpc);
    }else{
        ni->palette = (uint8_t*)malloc(ni->colors * ni->bpc);
    }
    return 0;
}


int
nile_ensure_bitplanes(nile_t *ni,int width, int height, int planes)
{
    int stride = (width-1)/8+1;
    int size = stride * height * planes;
    int i;
    ni->width = width;
    ni->height = height;
    ni->planes = planes;
    ni->bitplanes = malloc(ni->planes * sizeof(uint8_t*));
    ni->bitplanes[0] = malloc(size);
    for(i=1;i<ni->planes;++i)
        ni->bitplanes[i] = ni->bitplanes[i-1] + stride*ni->height;
    return 0;
}


int
nile_ensure_indexed(nile_t *ni,int width, int height, int colors){
    if(CHECK_MISMATCH(ni->width,width) ||
       CHECK_MISMATCH(ni->height,height))
        return -1;
    if(colors > 256)
        return -1;
    if(ni->colors){
        if(ni->colors < colors)
            return -1;
    }else{
        ni->colors = colors;
    }
    if(width>0)
        ni->width = width;
    if(height>0)
        ni->height = height;
    ni->indexed = malloc(ni->width*ni->height);
    return 0;
}


void
nile_drop_indexed(nile_t *ni)
{
    if(ni->indexed){
        free(ni->indexed);
        ni->indexed = NULL;
    }
}


int
nile_ensure_planar(nile_t *ni,int width, int height, int bpc)
{
    if(CHECK_MISMATCH(ni->width,width) ||
       CHECK_MISMATCH(ni->height,height))
        return -1;
    if(width>0)
        ni->width = width;
    if(height>0)
        ni->height = height;
    if(bpc>0)
        ni->bpc = bpc;
    ni->planar = malloc(ni->width * ni->height * ni->bpc);
    return 0;
}


void
nile_drop_planar(nile_t *ni)
{
    if(ni->planar){
        free(ni->planar);
        ni->planar = NULL;
    }
}


void
nile_drop_non_planar(nile_t *ni)
{
    if(ni->bitplanes){
        free(ni->bitplanes[0]);
        free(ni->bitplanes);
        ni->bitplanes = NULL;
    }
    nile_drop_indexed(ni);
}


static int
_get_stride(int bytes, int stride)
{
    if(stride>0)
        return stride;
    switch(stride){
    case 0:
    case NILE_STRIDE_ALIGN8:
        break;
    case NILE_STRIDE_ALIGN16:
        if(bytes&1)
            bytes += 1;
        break;
    case NILE_STRIDE_ALIGN32:
        if(bytes&3)
            bytes += 4 - (bytes&3);
        break;
    case NILE_STRIDE_ALIGN64:
        if(bytes&7)
            bytes += 8 - (bytes&7);
        break;
    }
    return bytes;
}


int
nile_get_bitplane_stride(const nile_t *ni)
{
    int bytes = (ni->width-1)/8+1;
    return _get_stride(bytes, ni->bitplane_stride);
}


int
nile_get_indexed_stride(const nile_t *ni)
{
    return _get_stride(ni->width, ni->indexed_stride);
}


int
nile_get_planar_stride(const nile_t *ni)
{
    return _get_stride(ni->width*ni->bpc, ni->planar_stride);
}


const uint8_t *
nile_planar_from_indexed(nile_t *ni){
    uint8_t *src = ni->indexed;
    int src_skip = nile_get_indexed_skip(ni);
    uint8_t *dst = ni->planar;
    int dst_skip = nile_get_planar_skip(ni);
    int y;
    for(y=0;y<ni->height;++y){
        int x;
        for(x=0;x<ni->width;++x){
            memcpy(dst, &ni->palette[(*src++)*ni->bpc], ni->bpc);
            dst += ni->bpc;
        }
        src += src_skip;
        dst += dst_skip;
    }
    return ni->planar;
}


double
nile_color_variance(const uint8_t *color1, nile_format_t format1, const uint8_t *color2, nile_format_t format2)
{
    if(format1 != format2){
        /* FIXME!! */
    }else{
        switch(NILE_COLORSPACE(format1)){
        case NILE_COLORSPACE_RGB:
            {
                double r = color1[0]-color2[0];
                double g = color1[1]-color2[1];
                double b = color1[2]-color2[2];
#if 1
                return 2.5*r*r + 4*g*g + 2.5*b*b;
#else
                return r*r + g*g + b*b;
#endif
            }
        default:
            break;
        }
    }
    return DBL_MAX;
}


int
nile_best_color_index(const uint8_t *p, int colors, int bpc, nile_format_t palette_format, const uint8_t *color, nile_format_t color_format)
{
    double best_value = DBL_MAX;
    int best_index = -1;
    int i;
    for(i=0;i<colors;++i){
        double value = nile_color_variance(p, palette_format, color, color_format);
        if(value < best_value){
            best_value = value;
            best_index = i;
        }
        p += bpc;
    }
    return best_index;
}


void
nile_copy_colors_rgb(uint8_t *dst, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_bpc, nile_format_t src_format, int count)
{
    int src_o_r;
    int src_o_g;
    int src_o_b;
    int src_o_a;
    int dst_o_r;
    int dst_o_g;
    int dst_o_b;
    int dst_o_a;
    if(src_bpc == dst_bpc && src_format == dst_format){
        memcpy(dst,src,count*src_bpc);
        return;
    }
    src_o_r = NILE_OFFSET_red(src_format);
    src_o_g = NILE_OFFSET_green(src_format);
    src_o_b = NILE_OFFSET_blue(src_format);
    src_o_a = NILE_OFFSET_alpha(src_format);
    dst_o_r = NILE_OFFSET_red(dst_format);
    dst_o_g = NILE_OFFSET_green(dst_format);
    dst_o_b = NILE_OFFSET_blue(dst_format);
    dst_o_a = NILE_OFFSET_alpha(dst_format);
    if(src_o_a != NILE_CHANNEL_MISSING && dst_o_a != NILE_CHANNEL_MISSING){
        while(count-->0){
            dst[dst_o_r] = src[src_o_r];
            dst[dst_o_g] = src[src_o_g];
            dst[dst_o_b] = src[src_o_b];
            dst[dst_o_a] = src[src_o_a];
            src += src_bpc;
            dst += dst_bpc;
        }
    }else if(dst_o_a != NILE_CHANNEL_MISSING){
        while(count-->0){
            dst[dst_o_r] = src[src_o_r];
            dst[dst_o_g] = src[src_o_g];
            dst[dst_o_b] = src[src_o_b];
            dst[dst_o_a] = 255;
            src += src_bpc;
            dst += dst_bpc;
        }
    }else{
        while(count-->0){
            dst[dst_o_r] = src[src_o_r];
            dst[dst_o_g] = src[src_o_g];
            dst[dst_o_b] = src[src_o_b];
            src += src_bpc;
            dst += dst_bpc;
        }
    }
}


#if 0
int nile_convert_colors(nile_t *ni,int bytes_per_color,nile_convert_colors_func_t func,void *userdata)
{
    uint8_t *palette = NULL;
    uint8_t *planar = NULL;
    int res;
    if(ni->palette){
        if(bytes_per_color != ni->bytes_per_color){
            palette = malloc(ni->colors * bytes_per_color);
            res = func(userdata,palette,ni->palette,ni->colors);
        }else{
            res = func(userdata,ni->palette,ni->palette,ni->colors);
        }
        if(res)
            goto error;
    }
    if(ni->planar){
        uint8_t *src = ni->planar;
        int src_stride = nile_get_planar_stride(ni);
        uint8_t *dst;
        int dst_stride;
        int y;
        if(bytes_per_color != ni->bytes_per_color){
            if(ni->planar_stride>0)
                ni->planar_stride = 0;
            dst_stride = _get_stride(ni->width * bytes_per_color, ni->planar_stride);
            dst = planar = malloc(dst_stride * ni->height);
        }else{
            dst = src;
            dst_stride = src_stride;
        }
        for(y=0;y<ni->height;++y){
            res = func(userdata,dst,src,ni->width);
            if(res)
                goto error;
            src += src_stride;
            dst += dst_stride;
        }
    }
    if(bytes_per_color != ni->bytes_per_color){
        ni->bytes_per_color = bytes_per_color;
        if(ni->palette){
            free(ni->palette);
            ni->palette = palette;
        }
        if(ni->planar){
            free(ni->planar);
            ni->planar = planar;
        }
    }
    return 0;
error:
    if(palette)
        free(palette);
    if(planar)
        free(planar);
    return res;
}
#endif
