/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef nile_h
#define nile_h

#include <math.h>                       /* sqrt() */
#include <endian.h>
#include <stddef.h>                     /* size_t */
#include <stdint.h>                     /* uint8_t */

#define NILE_MAX_BYTES_PER_COLOR 64
#define NILE_STRIDE_ALIGN8  -8
#define NILE_STRIDE_ALIGN16 -16
#define NILE_STRIDE_ALIGN32 -32
#define NILE_STRIDE_ALIGN64 -64

#define NILE_CHANNEL_MISSING 15
#define NILE_ENDIANESS     (1<<23)
#define NILE_PREMULTIPLIED (1<<29)
#define NILE_FLIPPED       (1<<0)

#define NILE_FORMAT_rgba8(r,g,b,a) (NILE_COLORSPACE_RGB<<26|NILE_CHANNELTYPE_INTEGER<<24|0<<20|(r)<<0|(g)<<4|(b)<<8|NILE_CHANNEL_MISSING<<12|(a)<<16)
#define NILE_FORMAT_rgb8(r,g,b) NILE_FORMAT_rgba8(r,g,b,NILE_CHANNEL_MISSING)
#define NILE_FORMAT_rgba16(r,g,b,a) (NILE_COLORSPACE_RGB<<26|NILE_CHANNELTYPE_INTEGER<<24|1<<20|(r)<<0|(g)<<4|(b)<<8|NILE_CHANNEL_MISSING<<12|(a)<<16)
#define NILE_FORMAT_rgb16(r,g,b) NILE_FORMAT_rgba16(r,g,b,NILE_CHANNEL_MISSING)


/*

31 30 29 28 27 26 25 24 23 22 21 20 19 18 17 16 15 14 13 12 11 10 09 08 07 06 05 04 03 02 01 00
         \      / \   /    \      / \         / \         / \         / \         / \         /
R     P     CS      C   E     CB       Alpha     Channel 3   Channel 2   Channel 1   Channel 0

R  = Reserved for custom formats
P  = Premultiplied
CS = Color space
C  = Encoding (integer or IEEE-754 floting point)
E  = Endianness (0 = big endian, 1 = little endian)
CB = Bytes per color channel (offset shifts): up to 128 bytes (1024 bits) per channel

*/

typedef enum{
    NILE_COLORSPACE_UNKNOWN = 0,
    NILE_COLORSPACE_RGB = 1,
    NILE_COLORSPACE_CMYK = 2,
    NILE_COLORSPACE_HSV = 3
}nile_colorspace_t;

#if 1
typedef enum{
    NILE_CHANNELTYPE_UNKNOWN = 0,
    NILE_CHANNELTYPE_INTEGER = 1,
    NILE_CHANNELTYPE_IEEE754 = 2
}nile_channeltype_t;
#else
typedef enum{
    NILE_CHANNELTYPE_UNKNOWN = 0,
    NILE_CHANNELTYPE_INT8 = 1,
    NILE_CHANNELTYPE_BINT16 = 2,
    NILE_CHANNELTYPE_LINT16 = 3,
    NILE_CHANNELTYPE_BINT32 = 4,
    NILE_CHANNELTYPE_LINT32 = 5,
    NILE_CHANNELTYPE_BINT32 = 4,
    NILE_CHANNELTYPE_LINT32 = 5,
    NILE_CHANNELTYPE_BINT64 = 6,
    NILE_CHANNELTYPE_LINT64 = 7,
    NILE_CHANNELTYPE_BFLOAT32 = 8,
    NILE_CHANNELTYPE_LFLOAT32 = 9,
    NILE_CHANNELTYPE_BFLOAT64 = 10,
    NILE_CHANNELTYPE_LFLOAT64 = 11,
    NILE_CHANNELTYPE_BFLOAT128 = 12,
    NILE_CHANNELTYPE_LFLOAT128 = 13,
}nile_channeltype_t;
#endif


#if 0
typedef union{
    int32_t id;
    struct{
#if __BYTE_ORDER == __BIG_ENDIAN
        union{
            struct{
                unsigned int r:4;
                unsigned int g:4;
                unsigned int b:4;
                unsigned int a:4;
            }
            struct{
        };
#else
        union{
        unsigned int a:4;
        unsigned int b:4;
        unsigned int g:4;
        unsigned int r:4;
#endif
    };
}nile_channel_order_t;
#endif

#define NILE_COLORSPACE(f) ((nile_colorspace_t)(((f)>>26)&7))
#define NILE_BYTES_PER_CHANNEL(f) (((f)>>20)&7)
#define NILE_OFFSET_CHANNEL(f,c) ((((f)>>((c)*4))&15)<<NILE_BYTES_PER_CHANNEL(f))
#define NILE_OFFSET_red(f)   NILE_OFFSET_CHANNEL(f,0)
#define NILE_OFFSET_green(f) NILE_OFFSET_CHANNEL(f,1)
#define NILE_OFFSET_blue(f)  NILE_OFFSET_CHANNEL(f,2)
#define NILE_OFFSET_alpha(f) NILE_OFFSET_CHANNEL(f,4)

#define NILE_HAS_ALPHA(f) (NILE_OFFSET_alpha(f)!=NILE_CHANNEL_MISSING)

typedef enum{
    //NILE_FORMAT_INVALID = -1,
    NILE_FORMAT_UNKNOWN = 0,
#if 1
    /* RGB */
    NILE_FORMAT_r8g8b8       = NILE_FORMAT_rgb8(0,1,2),
    NILE_FORMAT_x8r8g8b8     = NILE_FORMAT_rgb8(1,2,3),
    NILE_FORMAT_r8g8b8a8     = NILE_FORMAT_rgba8(0,1,2,3),
    NILE_FORMAT_a8r8g8b8     = NILE_FORMAT_rgba8(1,2,3,0),
    NILE_FORMAT_r16g16b16    = NILE_FORMAT_rgb16(0,1,3),
    NILE_FORMAT_r16g16b16a16 = NILE_FORMAT_rgba16(0,1,2,3),
    /* BGR */
    NILE_FORMAT_b8g8r8       = NILE_FORMAT_rgb8(2,1,0),
    NILE_FORMAT_x8b8g8r8     = NILE_FORMAT_rgb8(3,2,1),
    NILE_FORMAT_b8g8r8a8     = NILE_FORMAT_rgba8(2,1,0,3),
    NILE_FORMAT_a8b8g8r8     = NILE_FORMAT_rgba8(3,2,1,0),
    NILE_FORMAT_b16g16r16a16 = NILE_FORMAT_rgba16(2,1,0,3),
    /* gray */
    NILE_FORMAT_g8           = NILE_FORMAT_rgb8(0,0,0),
    NILE_FORMAT_g8a8         = NILE_FORMAT_rgba8(0,0,0,1),
    NILE_FORMAT_a8g8         = NILE_FORMAT_rgba8(1,1,1,0),
    NILE_FORMAT_g16          = NILE_FORMAT_rgb16(0,0,0),
    NILE_FORMAT_g16a16       = NILE_FORMAT_rgba16(0,0,0,1),
    NILE_FORMAT_a16g16       = NILE_FORMAT_rgba16(1,1,1,0)
#elif 1
    /* RGB */
    NILE_FORMAT_r8g8b8       = NILE_MAKE_FORMAT(0,1,2,NILE_CHANNEL_MISSING),
    NILE_FORMAT_x8r8g8b8     = NILE_MAKE_FORMAT(1,2,3,NILE_CHANNEL_MISSING),
    NILE_FORMAT_r8g8b8a8     = NILE_MAKE_FORMAT(0,1,2,3),
    NILE_FORMAT_a8r8g8b8     = NILE_MAKE_FORMAT(1,2,3,0),
    NILE_FORMAT_r16g16b16    = NILE_MAKE_FORMAT(0,2,4,NILE_CHANNEL_MISSING),
    NILE_FORMAT_r16g16b16a16 = NILE_MAKE_FORMAT(0,2,4,6),
    /* BGR */
    NILE_FORMAT_b8g8r8       = NILE_MAKE_FORMAT(2,1,0,NILE_CHANNEL_MISSING),
    NILE_FORMAT_x8b8g8r8     = NILE_MAKE_FORMAT(3,2,1,NILE_CHANNEL_MISSING),
    NILE_FORMAT_b8g8r8a8     = NILE_MAKE_FORMAT(2,1,0,3),
    NILE_FORMAT_a8b8g8r8     = NILE_MAKE_FORMAT(3,2,1,0),
    NILE_FORMAT_b16g16r16a16 = NILE_MAKE_FORMAT(4,2,0,6),
    /* gray */
    NILE_FORMAT_g            = NILE_MAKE_FORMAT(0,0,0,NILE_CHANNEL_MISSING),
    NILE_FORMAT_g8a8         = NILE_MAKE_FORMAT(0,0,0,1),
    NILE_FORMAT_a8g8         = NILE_MAKE_FORMAT(1,1,1,0)
#else
    NILE_FORMAT_G = 1,
    NILE_FORMAT_GA = 2,
    NILE_FORMAT_AG = 3,
    NILE_FORMAT_RGB = 4,
    NILE_FORMAT_BGR = 5,
    NILE_FORMAT_RGBA = 6,
    NILE_FORMAT_BGRA = 7,
    NILE_FORMAT_ARGB = 8,
    NILE_FORMAT_ABGR = 9,
    NILE_FORMAT_XRGB = 10,
    NILE_FORMAT_XBGR = 11
#endif
}nile_format_t;

typedef struct{
    int width;
    int height;
    nile_format_t format;
    int bpc;                            /** Bytes per color. */
    int colors;
    int planes;
    uint8_t *palette;
    uint8_t **bitplanes;
    uint8_t *indexed;
    uint8_t *planar;
    int bitplane_stride;
    int indexed_stride;
    int planar_stride;
}nile_t;

typedef int(*nile_convert_colors_func_t)(void *userdata,uint8_t *dst,const uint8_t *src,size_t count);

nile_t *nile_init(nile_t*);
void nile_done(nile_t*);
nile_t *nile_new(void);
void nile_free(nile_t*);

int nile_ensure_palette(nile_t *,int colors, int bytes_per_color);
void nile_drop_palette(nile_t*);
int nile_ensure_bitplanes(nile_t *,int width, int height, int planes);
int nile_ensure_indexed(nile_t *,int width, int height, int colors);
void nile_drop_indexed(nile_t*);
int nile_ensure_planar(nile_t *,int width, int height, int bytes_per_color);
void nile_drop_planar(nile_t*);
void nile_drop_non_planar(nile_t*);

int nile_get_bitplane_stride(const nile_t*);
int nile_get_indexed_stride(const nile_t*);
#define nile_get_indexed_skip(ni) (nile_get_indexed_stride(ni)-(ni)->width)
int nile_get_planar_stride(const nile_t*);
#define nile_get_planar_skip(ni) (nile_get_planar_stride(ni)-(ni)->width*(ni)->bpc)

const uint8_t *nile_planar_from_indexed(nile_t*);

double nile_color_variance(const uint8_t *color1, nile_format_t format1, const uint8_t *color2, nile_format_t format2);
#define nile_color_distance(c1,f1,c2,f2) sqrt(nile_color_variance(c1,f1,c2,f2))
int nile_best_color_index(const uint8_t *palette, int colors, int bpc, nile_format_t palette_format, const uint8_t *color, nile_format_t color_format);
void nile_copy_colors_rgb(uint8_t *dst, int dst_bpc, nile_format_t dst_format, const uint8_t *src, int src_bpc, nile_format_t src_format, int count);

//int nile_convert_colors(nile_t*,int bytes_per_color,nile_convert_colors_func_t,void *userdata);

#endif /* nile_h */
