/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */
#include <stdlib.h>                     /* malloc(), free() */

#include "pool.h"

#define MINSIZE 4096

#ifndef MAX
# define MAX(a,b) ((a)>(b)?(a):(b))
#endif


static void block_init(pool_block_t *b, size_t itemsize, size_t max){
    b->max = max;
    b->mem = malloc(max * itemsize);
    b->end = b->mem + max * itemsize;
    b->top = 0;
    b->count = 0;
    b->freelist = malloc(max * sizeof(void*));
    b->freelist_size = 0;
}


static void block_done(pool_block_t *b){
    free(b->mem);
    free(b->freelist);
}


pool_t *pool_init(pool_t *p,size_t itemsize, size_t minsize){
    p->itemsize = itemsize;
    block_init(&p->block[0],itemsize,MAX(MINSIZE,minsize));
    p->top = 0;
    p->current = 0;
    return p;
}


void pool_done(pool_t *p){
    int i;
    for(i=p->top;i>=0;--i)
        block_done(&p->block[i]);
}


void *pool_lease(pool_t *p){
    for(;;){
        pool_block_t *b = &p->block[p->current];
        if(b->count < b->max){
            /* There is room in the current block. */
            ++b->count;
            if(b->freelist_size > 0){
                return b->freelist[--b->freelist_size];
            }else{
                return &b->mem[b->top++ * p->itemsize];
            }
        }
        ++p->current;
        if(p->current > p->top){
            /* Allocate a new block. */
            ++p->top;
            block_init(&p->block[p->top], p->itemsize, p->block[0].max << p->top);
        }
    }
}


void pool_return(pool_t *p,void *mem){
    int i;
    for(i=p->current;i>=0;--i){
        pool_block_t *b = &p->block[i];
        if(mem >= (void*)b->mem && mem < b->end){
            /* The memory belongs to this block. */
            --b->count;
            if(mem == b->mem + b->top - p->itemsize){
                /* it was the top item */
                --b->top;
            }else{
                b->freelist[b->freelist_size++] = mem;
            }
            if(b->count <= 0){
                /* The whole block is free. */
                b->top = 0;
                b->freelist_size = 0;

                if(i==p->current){
                    while(p->current > 0){
                        --p->current;
                        if(p->top - p->current >= 2)
                            block_done(&p->block[p->top--]);
                        b = &p->block[p->current];
                        if(b->count > 0)
                            break;
                    }
                }
            }
            return;
        }
    }
}
